@@ -36,16 +36,57 @@
 #include "connection.h"
 #include "utils.h"
 
+static void uh_server_free_conns(struct list_head *head)
+{
+    struct uh_connection_internal *pos, *n;
+
+    list_for_each_entry_safe(pos, n, head, list) {
+        conn_free(pos);
+    }
+}
+
+static void uh_server_free_handlers(struct list_head *head)
+{
+    struct uh_path_handler *pos, *n;
+
+    list_for_each_entry_safe(pos, n, head, list) {
+        list_del(&pos->list);
+        free(pos);
+    }
+}
+
+static void uh_server_free_plugins(struct list_head *head)
+{
+#ifdef HAVE_DLOPEN
+    struct uh_plugin *pos, *n;
+
+    list_for_each_entry_safe(pos, n, head, list) {
+        list_del(&pos->list);
+        dlclose(pos->dlh);
+        free(pos);
+    }
+#endif
+}
+
+static void uh_server_free_listeners(struct list_head *head)
+{
+    struct uh_listener *pos, *n;
+
+    list_for_each_entry_safe(pos, n, head, list) {
+        ev_io_stop(pos->srv->loop, &pos->ior);
+
+        list_del(&pos->list);
+
+        if (pos->sock > 0)
+            close(pos->sock);
+
+        free(pos);
+    }
+}
 
 static void uh_server_free(struct uh_server *srv)
 {
     struct uh_server_internal *srvi = (struct uh_server_internal *)srv;
-    struct uh_connection_internal *conn = srvi->conns;
-    struct uh_path_handler *h = srvi->handlers;
-    struct uh_listener *l = srvi->listeners;
-#ifdef HAVE_DLOPEN
-    struct uh_plugin *p = srvi->plugins;
-#endif
 
     if (srvi->docroot)
         free(srvi->docroot);
@@ -53,38 +94,10 @@ static void uh_server_free(struct uh_server *srv)
     if (srvi->index_page)
         free(srvi->index_page);
 
-    while (conn) {
-        struct uh_connection_internal *next = conn->next;
-        conn_free(conn);
-        conn = next;
-    }
-
-    while (h) {
-        struct uh_path_handler *temp = h;
-        h = h->next;
-        free(temp);
-    }
-
-    while (l) {
-        struct uh_listener *temp = l;
-
-        ev_io_stop(srvi->loop, &l->ior);
-
-        if (l->sock > 0)
-            close(l->sock);
-
-        l = l->next;
-        free(temp);
-    }
-
-#ifdef HAVE_DLOPEN
-    while (p) {
-        struct uh_plugin *temp = p;
-        dlclose(p->dlh);
-        p = p->next;
-        free(temp);
-    }
-#endif
+    uh_server_free_conns(&srvi->conns);
+    uh_server_free_handlers(&srvi->handlers);
+    uh_server_free_plugins(&srvi->plugins);
+    uh_server_free_listeners(&srvi->listeners);
 
 #ifdef SSL_SUPPORT
     ssl_context_free(srvi->ssl_ctx);
@@ -94,8 +107,6 @@ static void uh_server_free(struct uh_server *srv)
 static void uh_accept_cb(struct ev_loop *loop, struct ev_io *w, int revents)
 {
     struct uh_listener *l = container_of(w, struct uh_listener, ior);
-    struct uh_server_internal *srv = l->srv;
-    struct uh_connection_internal *conn;
     union {
         struct sockaddr     sa;
         struct sockaddr_in  sin;
@@ -113,12 +124,12 @@ static void uh_accept_cb(struct ev_loop *loop, struct ev_io *w, int revents)
         return;
     }
 
-    log_debug("New Connection from: %s %d\n",
-            saddr2str(&addr.sa, addr_str, sizeof(addr_str), &port), port);
+    log_info("New Connection from %s %d\n", addr_str,
+            (saddr2str(&addr.sa, addr_str, sizeof(addr_str), &port) ? port : 0));
 
     if (l->ssl) {
 #ifdef SSL_SUPPORT
-        if (!srv->ssl_ctx) {
+        if (!l->srv->ssl_ctx) {
             log_err("SSL not initialized\n");
             close(sock);
             return;
@@ -130,18 +141,7 @@ static void uh_accept_cb(struct ev_loop *loop, struct ev_io *w, int revents)
 #endif
     }
 
-    conn = uh_new_connection(l, sock, &addr.sa);
-    if (!conn)
-        return;
-
-    if (!srv->conns) {
-        srv->conns = conn;
-        return;
-    }
-
-    conn->next = srv->conns;
-    srv->conns->prev = conn;
-    srv->conns = conn;
+    uh_new_connection(l, sock, &addr.sa);
 }
 
 struct uh_server *uh_server_new(struct ev_loop *loop)
@@ -219,14 +219,21 @@ static int uh_load_plugin(struct uh_server *srv, const char *path)
 
     p->h = h;
     p->dlh = dlh;
+    p->path = h->path;
+    p->len = strlen(h->path);
 
-    if (!srvi->plugins) {
-        srvi->plugins = p;
-        return 0;
+    if (h->path[0] == '^') {
+        p->flags |= UH_PATH_MATCH_START;
+        p->len--;
+        p->path++;
     }
 
-    p->next = srvi->plugins;
-    srvi->plugins = p;
+    if (p->path[p->len - 1] == '$') {
+        p->flags |= UH_PATH_MATCH_END;
+        p->len--;
+    }
+
+    list_add(&p->list, &srvi->plugins);
 
     return 0;
 #else
@@ -238,7 +245,20 @@ static int uh_load_plugin(struct uh_server *srv, const char *path)
 static int uh_add_path_handler(struct uh_server *srv, const char *path, uh_path_handler_prototype handler)
 {
     struct uh_server_internal *srvi = (struct uh_server_internal *)srv;
+    int path_len = strlen(path);
     struct uh_path_handler *h;
+    uint8_t flags = 0;
+
+    if (path[0] == '^') {
+        flags |= UH_PATH_MATCH_START;
+        path_len--;
+        path++;
+    }
+
+    if (path[path_len - 1] == '$') {
+        flags |= UH_PATH_MATCH_END;
+        path_len--;
+    }
 
     h = calloc(1, sizeof(struct uh_path_handler) + strlen(path) + 1);
     if (!h) {
@@ -247,15 +267,12 @@ static int uh_add_path_handler(struct uh_server *srv, const char *path, uh_path_
     }
 
     h->handler = handler;
-    strcpy(h->path, path);
+    h->flags = flags;
+    h->len = path_len;
 
-    if (!srvi->handlers) {
-        srvi->handlers = h;
-        return 0;
-    }
+    strncpy(h->path, path, path_len);
 
-    h->next = srvi->handlers;
-    srvi->handlers = h;
+    list_add(&h->list, &srvi->handlers);
 
     return 0;
 }
@@ -418,12 +435,7 @@ static int uh_server_listen(struct uh_server *srv, const char *addr, bool ssl)
         ev_io_init(&l->ior, uh_accept_cb, sock, EV_READ);
         ev_io_start(srvi->loop, &l->ior);
 
-        if (!srvi->listeners) {
-            srvi->listeners = l;
-        } else {
-            l->next = srvi->listeners;
-            srvi->listeners = l;
-        }
+        list_add(&l->list, &srvi->listeners);
 
         if (p->ai_family == AF_INET) {
             struct sockaddr_in *ina = (struct sockaddr_in *)p->ai_addr;
@@ -455,6 +467,11 @@ void uh_server_init(struct uh_server *srv, struct ev_loop *loop)
 
     memset(srvi, 0, sizeof(struct uh_server_internal));
 
+    INIT_LIST_HEAD(&srvi->listeners);
+    INIT_LIST_HEAD(&srvi->handlers);
+    INIT_LIST_HEAD(&srvi->plugins);
+    INIT_LIST_HEAD(&srvi->conns);
+
     srvi->loop = loop ? loop : EV_DEFAULT;
 
     srv->get_loop = uh_get_loop;
